(******************************************************************************
 * Copyright (c) 2012-2013, Toyama&Aoto Laboratory, Tohoku University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(******************************************************************************
 * file: rwtools/commands/tpdb.yacc
 * description: parser for TPDB (Termination Problem Data Base) format
 * author: AOTO Takahito
 * 
 * $Id: tpdb.yacc,v 1.9 2013/06/09 11:56:50 aoto Exp $
 ******************************************************************************)

local 
    open Term
    structure L = List
    structure LU = ListUtil
    structure PU = PrintUtil
in
datatype tpdbExpr = VAR of string list 
      		  | RULES of (term * term) option list
		  | THEORY of (string * fun_key list) list

datatype tpdb_thdecl = THEORIES of string * string list 
                     | EQUATIONS of (term * term) list

datatype tpdb_cond = ORIENT of (term * term)
                   | JOIN of (term * term)

datatype tpdb_rule = NORMAL of (term * term * tpdb_cond list)
                   | RELATIVE of (term * term * tpdb_cond list)

datatype tpdb_strat = STRAT_INNERMOST
                    | STRAT_OUTERMOST
                    | STRAT_CS of (fun_key * int list) list

datatype tpdb_decl = VAR of Atom.atom list 
		   | THEORY of tpdb_thdecl list
                   | RULES of tpdb_rule list 
                   | STRATEGY of tpdb_strat
                   | REST of string list

exception TpdbError of string

fun getCategory f xss = let val yss = List.mapPartial f xss 
			in if null yss 
			   then NONE
			   else SOME (LU.mapAppend (fn x=>x) yss)
			end
val getVAR = getCategory (fn top => case top of VAR xs => SOME xs | _ => NONE) 
val getRULES = getCategory (fn top => case top of RULES xs => SOME xs | _ => NONE)
val getTHEORY = getCategory (fn top => case top of THEORY xs => SOME xs | _ => NONE)
val getREST = getCategory (fn top => case top of REST xs => SOME xs | _ => NONE)

(* for the stragegy, only the fist one is considerd *)
fun getSTRATEGY (top::rest) = (case top of STRATEGY xs => SOME xs | _ => getSTRATEGY rest)
  | getSTRATEGY [] = NONE
						  
fun makeTrs vars rules = 
    let fun transTerm (t as (Fun (y,[],_))) = 
	    if ListUtil.member' Fun.equal y vars
	    then (Var ((y,0), Sort.null)) 
	    else t
	  | transTerm (Fun (y,ys,_)) = Fun (y, List.map transTerm ys, Sort.null)
    in
	List.map (fn (l,r) => (transTerm l, transTerm r)) rules
    end

local
    exception NonSupportedRule
in
fun checkRules xs ys = 
    let	fun getNormal ys = List.mapPartial 
                              (fn y => case y of 
                                NORMAL (l,r,xs) => 
    if null xs
                          then SOME (l,r)
                          else raise NonSupportedRule
                                | _ => NONE) ys
       fun getRelative ys = List.mapPartial 
                            (fn y => case y of 
                                RELATIVE (l,r,xs) => 
                          if null xs
                          then SOME (l,r)
                          else raise NonSupportedRule
                                | _ => NONE) ys
    in (makeTrs xs (getNormal ys),
        makeTrs xs (getRelative ys))
    end
    handle NonSupportedRule => 
                 (PU.printStdErr "CONDITIONAL rules are not supported.\n"; 
                  raise (TpdbError "unsupported format"))
end

fun getRules decls =
    case (getVAR decls, getRULES decls) of
	(SOME xs, SOME ys) => checkRules xs ys
      | (NONE, SOME ys) => checkRules [] ys
      | (SOME xs,NONE) => checkRules xs []
      | (NONE,NONE) => checkRules [] []

fun checkDecl decls = 
    if isSome (getTHEORY decls)
    then (PU.printStdErr "THEORY is not supported.\n"; 
          raise (TpdbError "unsupported format"))
    else if isSome (getSTRATEGY decls)
    then (PU.printStdErr "STRATEGY is not supported.\n"; 
          raise (TpdbError "unsupported format"))
    else case getREST decls of
	     SOME xs => (PU.print ("(ignored inputs)" ^  
                         (ListUtil.toStringSpace (fn x=>x) xs) ^ "\n");
			 getRules decls)
	   | NONE => getRules decls

end

(* 桼 *)
%%
%name		Tpdb
%term           ID of string
              | LPAREN
              | RPAREN
              | ARROW
              | EQUAL
              | ARROWEQ
              | JOIN
              | BAR
              | COMMA
              | DQUOTE
              | CONTEXTSENSITIVE
              | EQUATIONS
              | INNERMOST
              | OUTERMOST
              | RULES
              | STRATEGY
              | THEORY
              | VAR 
              | EOF 
%nonterm	XstartX of ((Term.term * Term.term) list * (Term.term * Term.term) list)
              | XspecX of tpdb_decl list
              | XdeclX of tpdb_decl
              | XanylistX of string list
              | XstringX of string list
              | XidlistX of string list
              | XlistofthdeclX of tpdb_thdecl list
              | XthdeclX of tpdb_thdecl
              | XeqlistX of (Term.term * Term.term) list
              | XequationX of Term.term * Term.term
              | XlistofrulesX of tpdb_rule list
              | XruleX of tpdb_rule
              | XcondlistX of tpdb_cond list
              | XcondX of tpdb_cond
              | XtermX of Term.term
              | XtermlistX of Term.term list
              | XstrategydeclX of tpdb_strat
              | XcsstratlistX of (Fun.ord_key * int list) list
              | XintlistX of int list
              | XintX of int
%pos	   	int
%eop		EOF
%noshift	EOF

%%
XstartX:      XspecX                       (checkDecl (XspecX))
XspecX:       LPAREN XdeclX RPAREN XspecX  (XdeclX::XspecX)
            |                              ([])
XdeclX:       VAR XidlistX                 (VAR (List.map Atom.atom (XidlistX)))
            | THEORY XlistofthdeclX        (THEORY XlistofthdeclX)
            | RULES XlistofrulesX          (RULES XlistofrulesX)
            | STRATEGY XstrategydeclX      (STRATEGY XstrategydeclX)
            | ID XanylistX                 (REST (ID::XanylistX))
XanylistX:    ID XanylistX                 (ID::XanylistX)
            | XstringX XanylistX           (XstringX @ XanylistX)
            | LPAREN XanylistX RPAREN XanylistX  (("("::XanylistX1) @ (")"::XanylistX2))
            | JOIN XanylistX               ("-><-"::XanylistX)
            | ARROW XanylistX              ("->"::XanylistX)
            | EQUAL XanylistX              ("="::XanylistX)
            | BAR XanylistX                ("|"::XanylistX)
            | ARROWEQ XanylistX            ("->="::XanylistX)
            | COMMA XanylistX              (","::XanylistX)
            | VAR XanylistX               ("VAR"::XanylistX)
            | RULES XanylistX              ("RULES"::XanylistX)
            | CONTEXTSENSITIVE XanylistX   ("CONTEXTSENSITIVE"::XanylistX)
            | EQUATIONS XanylistX          ("EQUATIONS"::XanylistX)
            | INNERMOST XanylistX          ("INNERMOST"::XanylistX)
            | OUTERMOST XanylistX          ("OUTERMOST"::XanylistX)
            | STRATEGY XanylistX           ("STRATEGY"::XanylistX)
            | THEORY XanylistX             ("THEORY"::XanylistX)
            |                              ([])
XstringX:     DQUOTE XidlistX DQUOTE       (XidlistX)
XidlistX:     ID XidlistX                  (ID::XidlistX)
            |                              ([])
XlistofthdeclX: LPAREN XthdeclX RPAREN XlistofthdeclX (XthdeclX::XlistofthdeclX)
            |                              ([])
XthdeclX:     ID XidlistX                  (THEORIES (ID,XidlistX))
            | EQUATIONS XeqlistX           (EQUATIONS XeqlistX)
XeqlistX:     XequationX XeqlistX          (XequationX::XeqlistX)
            |                              ([])
XequationX:   XtermX EQUAL XtermX            ((XtermX1,XtermX2))
XlistofrulesX:  XruleX  XlistofrulesX      (XruleX::XlistofrulesX)
            |                              ([])
XruleX:       XtermX ARROW XtermX          (NORMAL (XtermX1,XtermX2,[]))
            | XtermX ARROW XtermX BAR XcondlistX (NORMAL(XtermX1,XtermX2,XcondlistX))
            | XtermX ARROWEQ XtermX        (RELATIVE (XtermX1,XtermX2,[]))
            | XtermX ARROWEQ XtermX BAR XcondlistX  (RELATIVE (XtermX1,XtermX2,XcondlistX))
XcondlistX:   XcondX                       ([XcondX])
            | XcondX COMMA XcondlistX      (XcondX::XcondlistX)
XcondX:       XtermX ARROW XtermX          (ORIENT (XtermX1,XtermX2))
            | XtermX JOIN XtermX           (JOIN (XtermX1,XtermX2))
XtermX:       ID                           (Term.Fun (Atom.atom ID,[],Sort.null))
            | ID LPAREN RPAREN             (Term.Fun (Atom.atom ID,[],Sort.null))
            | ID LPAREN XtermlistX RPAREN  (Term.Fun (Atom.atom ID,XtermlistX,Sort.null))
XtermlistX:   XtermX COMMA XtermlistX      (XtermX::XtermlistX)
            | XtermX 			   ([XtermX])
XstrategydeclX: INNERMOST                  (STRAT_INNERMOST)
            | OUTERMOST                    (STRAT_OUTERMOST)
            | CONTEXTSENSITIVE XcsstratlistX (STRAT_CS XcsstratlistX)
XcsstratlistX:  LPAREN ID  XintlistX RPAREN XcsstratlistX  ((Atom.atom ID,XintlistX)::XcsstratlistX)
            |                              ([])
XintlistX:    XintX XintlistX              (XintX::XintlistX) 
            |                              ([])
XintX:        ID                           (valOf (Int.fromString ID))
